// -------------------------------------------------------
//
// Author: Thomas Lenart    
// Email:  thomas.lenart@es.lth.se
// Date:   March 2003
//
// -------------------------------------------------------

#include "main.h"
#include "memhandler.h"
#include "bitstream.h"

/*
	Example of reading 5 bits starting at bit 30

	The stream fetches values from a 32 bit buffer

    |31|30|29|28|27|26|25|24|23|22|21|20|19|18|17|16|15|14|13|12|11|10|09|08|07|06|05|04|03|02|01|00|
	-------------------------------------------------------------------------------------------------
0:  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |x1|x2|
1:  |x3|x4|x5|  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
2:  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |

	Return value : x1(MSB) x2 x3 x4 x5(LSB)

*/

CT_Bitstream *bitstream_create(unsigned char *data, int size)
{
	CT_Bitstream *stream;

	stream = (CT_Bitstream*)malloc(sizeof(CT_Bitstream));

	stream->data = data;
	stream->size = size;
	stream->modulus = size;
	stream->bit_position = 0;

	return stream;
}

void bitstream_delete(CT_Bitstream *stream)
{
	free(stream);
}

void bitstream_init(CT_Bitstream *stream, unsigned char *data, int size)
{
	stream->data = data;
	stream->size = size;
	stream->modulus = size;
	stream->bit_position = 0;
}

int bitstream_read(CT_Bitstream *stream, int bits)
{
	unsigned int	value, next, mask;
	int				offset;

	value =	stream->data[stream->bit_position >> 3];					// get the first row
	mask  = (1 << bits) - 1;											// calculate mask

	offset = 8 - (stream->bit_position & 0x07) - bits;					// number of steps to move the first row
	if (offset > 0)														// if in the middle of the word
		value >>= offset;												// move it to the LSB of value
	else
		value <<= -offset;												// shift it to the left

	if (offset < 0) {													// the line is wrapped
		next = stream->data[(stream->bit_position >> 3) + 1];			// get the second row
		next >>= 8 + offset;											// get MSB bits
		value += next & ((1 << (-offset)) - 1);							// add to value
	}

	value &= mask;														// mask the output value

	stream->bit_position += bits;

	return value;
}

int bitstream_read_mod(CT_Bitstream *stream, int bits)
{
	unsigned int	value, next, mask;
	int				offset;

	value =	stream->data[(stream->bit_position >> 3) & (stream->modulus-1)];		// get the first row

	mask  = (1 << bits) - 1;														// calculate mask

	offset = 8 - (stream->bit_position & 0x07) - bits;								// number of steps to move the first row
	if (offset > 0)																	// if in the middle of the word
		value >>= offset;															// move it to the LSB of value
	else
		value <<= -offset;															// shift it to the left

	if (offset < 0) {																// the line is wrapped
		next = stream->data[((stream->bit_position >> 3) + 1) & (stream->modulus-1)];	// get the second row
		next >>= 8 + offset;														// get MSB bits
		value += next & ((1 << (-offset)) - 1);										// add to value
	}

	value &= mask;																	// mask the output value

	stream->bit_position += bits;

	return value;
}

//--------------------
// Inline functions
//--------------------

int bitstream_read_16_mod(CT_Bitstream *stream, int bits)
{
	int value;
	
	value  = bitstream_read_mod(stream, 8) << (bits-8);
	value += bitstream_read_mod(stream, bits-8);

	return value;
}

int bitstream_read_16(CT_Bitstream *stream, int bits)
{
	int value;
	
	value  = bitstream_read(stream, 8) << (bits-8);
	value += bitstream_read(stream, bits-8);

	return value;
}

int bitstream_read_1(CT_Bitstream *stream)
{
	unsigned int value;

//	#ifdef _WIN32
		value = stream->data[stream->bit_position >> 3];							// get 32 bit word
		value = (value >> (7-(stream->bit_position++ & 0x07))) & 1;					// move bit to LSB
/*	#else
		unsigned char* data = stream->data;
		unsigned int*  bit_position = &stream->bit_position;
		unsigned int   one, shift, bit_pos;

		__asm {
			LDR		bit_pos, [bit_position]
			ADD		data, data, bit_pos, LSR #3
			ADD		bit_pos, bit_pos, #1
			STR		bit_pos, [bit_position]
			LDRB	value, [data]			
			
			AND		shift, bit_pos, #7
			RSB		shift, shift, #7

			MOV		one, #1
			AND		value, one, value, LSR shift
		}
	#endif*/

	return value;
}

int bitstream_read_1_mod(CT_Bitstream *stream)
{
	unsigned int value;

	value = stream->data[(stream->bit_position >> 3) & (stream->modulus-1)];	// get 32 bit word
	value = (value >> (7-(stream->bit_position++ & 0x07))) & 1;					// move bit to LSB

	return value;
}

int bitstream_eos(CT_Bitstream *stream)
{
	if (stream->bit_position > (stream->size << 3))								// if end of stream
		return 1;
		
	return 0;
}

int bitstream_pos(CT_Bitstream *stream)								// get bitstream position
{
	return (stream->bit_position);
}

void bitstream_set_pos(CT_Bitstream *stream, int position)				// set bitstream position
{
	stream->bit_position = position;
}
